#ifndef CIRCULAR_CONVEX_FUNCTION_HH
#define CIRCULAR_CONVEX_FUNCTION_HH

#include <cmath>
#include <dune/fufem/interval.hh>

namespace Dune {
template <class MatrixType, class VectorType> class CircularConvexFunction {
public:
  CircularConvexFunction(MatrixType const &A, VectorType const &b,
                         VectorType const &x, VectorType const &dir)
      : A(A),
        b(b),
        x(x),
        dir(dir),
        xnorm(x.two_norm()),
        dnorm(dir.two_norm()) {}

  double quadraticPart() const { return 0; }

  double linearPart() const { return 0; }

  void subDiff(double m, Interval<double> &D) const {
    VectorType x;
    cartesian(m, x);
    VectorType t;
    tangent(m, t);

    VectorType tmp;
    A.mv(x, tmp);          //  Ax
    tmp -= b;              //  Ax - b
    D[0] = D[1] = tmp * t; // <Ax - b,t>
  }

  void domain(Interval<double> &domain) const {
    domain[0] = -2 * M_PI;
    domain[1] = 2 * M_PI;
  }

  void cartesian(double m, VectorType &y) const {
    y = 0;
    y.axpy(std::cos(m), x);
    y.axpy(std::sin(m) * xnorm / dnorm, dir);
  }

private:
  MatrixType const &A;
  VectorType const &b;
  VectorType const &x;
  VectorType const &dir;

  double const dnorm;
  double const xnorm;

  /* If x and d were normalised, we would have

       cartesian(a) =  cos(a) * x + sin(a) * d and
       tangent(a)   = -sin(a) * x + cos(a) * d.

     Since we x and d are not normalised and the return of
     cartesian() is fixed, we scale the tangent.
  */
  void tangent(double m, VectorType &y) const {
    y = 0;
    y.axpy(-std::sin(m) * dnorm, x);
    y.axpy(std::cos(m) * xnorm, dir);
  }
};
}

#endif
